/**
* tca.c - P2PWNC Trusted Central Authority (TCA) module
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "tca.h"

pthread_t team_handler_thread;

void term_handler(int signum) {
	int i = 0;
	for (i = 0; i<max_threads; i++) {
		cancel_thread(i);
	}
#ifdef FIXED_IDS
	pthread_cancel(team_handler_thread);
#endif
	cleanup_ossl_threads();
	thread_pool_cleanup();
	exit(0);
}

//initializes message queue
int init_mq() {
	fprintf(stderr, "init_mq: Initializing message queue\n");
	mq_id = msgget(mq_key, IPC_CREAT | 0400);
	if (mq_id == -1) {
		fprintf(stderr, "init_mq: Error getting message queue\n");
	}
	return mq_id;
}

//handles team additions/deletions
int handle_team_msg() {
	MSG_BUF *msg_to_receive = NULL;
	PWNC_PUBKEY *pubkey = NULL;

	ALLOC_AND_CLEAN(msg_to_receive, MAX_MSG_LEN, MSG_BUF*); //allocate memory for msgbuf

	if (msgrcv(mq_id, msg_to_receive, MAX_MSG_LEN, 0, 0) == -1) {
		fprintf(stderr, "rcv_cmd: Error receiving message from queue\n");
		SAFE_FREE(msg_to_receive);
		return -1;
	}
	else {
		fprintf(stderr, "rcv_cmd: Received message from queue\n");
		ALLOC_AND_CLEAN(pubkey, sizeof(PWNC_PUBKEY), PWNC_PUBKEY*);
		memcpy( &(pubkey->algorithm), msg_to_receive->mdata, sizeof(unsigned short) );
		memcpy( &(pubkey->bits), msg_to_receive->mdata + sizeof(unsigned short), sizeof(unsigned short) );
		memcpy( &(pubkey->datalen), msg_to_receive->mdata + 2*sizeof(unsigned short), sizeof(unsigned int) );
		ALLOC_AND_CLEAN(pubkey->data, pubkey->datalen, unsigned char*);
		memcpy( pubkey->data, msg_to_receive->mdata + 2*sizeof(unsigned short) + sizeof(unsigned int), pubkey->datalen);

		if (msg_to_receive->mtype == MTYPE_ADD_TEAM) {
			//add specified team
			fprintf(stderr, "rcv_cmd: Adding team\n");
			rdb_add_team(pubkey);
			SAFE_FREE(msg_to_receive);
			return 0;
		}
		else
		if (msg_to_receive->mtype == MTYPE_DEL_TEAM) {
			//remove team
			//TODO: rdb_remove_team(to_be_handled)
			fprintf(stderr, "rcv_cmd: Removing teams has not yet been implemented\n");
			SAFE_FREE(msg_to_receive);
			return 0;
		}
		else {
			SAFE_FREE(msg_to_receive);
			return -1;
		}
	}
	SAFE_FREE(msg_to_receive);
	return 0;
}

//Scans the teams_path directory and loads the pubkeys of the teams that reside there
void load_teams(const char *teams_path) {
	int n = 0;
	int i = 0;
	struct dirent **namelist;
	PWNC_PUBKEY *team_key = NULL;
	unsigned char key_path[256];

	n = scandir(teams_path, &namelist, 0, alphasort);
	if (n < 0) {
		fprintf(stderr, "load_teams: scandir failed\n");
		return;
	}

	for (i = 0; i<n; i++) {
		if (!strcmp(namelist[i]->d_name, ".") || !strcmp(namelist[i]->d_name, "..") ) {
			continue;
		}
		memset(&key_path, 0, 256);
		sprintf(key_path, "%s/%s", teams_path, namelist[i]->d_name);
		fprintf(stderr, "Loading: %s\n", key_path);
		if (namelist[i]) {
			free(namelist[i]);
		}
		team_key = read_key_file(key_path);
		if (!team_key) {
			fprintf(stderr, "load_teams: Could not read key \"%s\"\n", key_path);
		}
		else {
			rdb_add_team(team_key);
		}
	}
	free(namelist);
}

void team_handler() {

	if (block_signals()) {
		fprintf(stderr, "Warning: Could not set signal mask\n");
	}
	init_mq();
	while (1) {
		handle_team_msg();
	}
}

//thread func that handles requests
void request_handler(THREAD_INFO *ti) {
	unsigned char *text = NULL; //text message
	int mtype = 0; //message type
	PWNC_RCPT *rcpt = NULL;
	PWNC_PUBKEY *provider = NULL;
	PWNC_PUBKEY *consumer = NULL;
	struct timeval delay; //delay
	int select_status = 0; //used in "select" calls
	int ecode = 0; //error code or 1(=OK)
	fd_set fs;

	if (block_signals()) {
		fprintf(stderr, "Warning: Could not set signal mask\n");
	}

	//get thread info
	int sockfd = ti->client_socket;
	int tindex = ti->thread_index;
	SAFE_FREE(ti);

	struct sockaddr_in client_addr;
	int addr_len = sizeof(client_addr);
	memset(&client_addr, 0, addr_len);

	//get client address from socket
	if (getpeername(sockfd, (struct sockaddr*)&client_addr, (socklen_t*)&addr_len) == -1) {
		fprintf(stderr, "request_handler: Error extracting addr info\n");
		exit_client_thread(tindex);
		return;
	}

	//read message from socket
	FD_ZERO(&fs);
	FD_SET(sockfd, &fs);
	delay.tv_sec = 20;//timeout: 5 seconds
	delay.tv_usec = 0;
	if ( (select_status = select(sockfd + 1, &fs, NULL, NULL, &delay)) > 0) {
		text = read_message(sockfd, &mtype);
		fprintf(stderr, "%s\n", text);
	}
	else {
		if (sockfd) {
			close(sockfd);
		}
		SAFE_FREE(text);
		//pthread_exit(NULL);
		exit_client_thread(tindex);
	}

	fprintf(stderr, "TCA received message %s\n", ( (mtype == PWNC_MSG_QUER)?"QUER":"RCPT") );
	switch (mtype) {
		case PWNC_MSG_QUER: {
			fprintf(stderr, "PWNC_TCA: Received QUER message\n%s\n", text);
			ALLOC_AND_CLEAN(provider, sizeof(PWNC_PUBKEY), PWNC_PUBKEY*);
			ALLOC_AND_CLEAN(consumer, sizeof(PWNC_PUBKEY), PWNC_PUBKEY*);
			if (!parse_query_msg(text, provider, consumer) && rdb_judge(consumer, provider)) {
				SAFE_FREE(text);
				//ok, access should be granted..
				text = construct_qrsp(PWNC_ACTION_GRANT);
			}
			else {
				SAFE_FREE(text);
				//access should not be granted or some other error occured
				text = construct_qrsp(PWNC_ACTION_FORBID);
			}

			SAFE_FREE_KEY(provider);
			SAFE_FREE_KEY(consumer);

			//sendQRSP message
			if (!write(sockfd, text, strlen(text))) {
				//error writing to socket
			}

			if (sockfd) {
				close(sockfd);
			}

			SAFE_FREE(text);
			//pthread_exit(NULL);
			exit_client_thread(tindex);
			break;
		}

		case PWNC_MSG_RCPT: {
			fprintf(stderr, "PWNC_TCA: Received RCPT message:\n%s\n", text);
			rcpt  = text_2_rcpt(text);
			if (!rcpt) {
				SAFE_FREE(text);
				//"unknown/other error" message
				text = construct_rrsp(PWNC_ERR_OTHER);
			}
			else {
#ifdef FIXED_IDS
				//if FIXED_IDS is set, first perform team lookup
				//if teams exist, proceed, else RRSP(err)
				if (	!rdb_team_exists(rcpt->ProviderPubKey) ||
					!rdb_team_exists(rcpt->ConsumerCert->TeamPubKey) ) {
					//some of the teams is unknown --> do nothing
					text = construct_rrsp(PWNC_ERR_TEAM_LOOKUP);
				}
				else {
					//else, both teams are known, so insert receipt
#endif
					ecode = insert_receipt(rcpt, 1); //TODO: check arg --> vlepe team_server
					text = construct_rrsp(ecode);//OK message
#ifdef FIXED_IDS
				}
#endif
			}
			if (!write(sockfd, text, strlen(text))) {
				//error writing to socket
			}

			if (sockfd) {
				close(sockfd);
			}
			SAFE_FREE(text);
			//pthread_exit(NULL);
			exit_client_thread(tindex);
		}
		default: {
			//TODO: send generic error message
			SAFE_FREE(text);
			if (sockfd) {
				close(sockfd);
			}
			//pthread_exit(NULL);
			exit_client_thread(tindex);
			break;
		}
	}
	//pthread_exit(NULL);
	exit_client_thread(tindex);
}


//##################################################################

void print_usage() {
	fprintf(stderr, "\nUsage: tca ");
	fprintf(stderr, "[-h/--help] ");
	fprintf(stderr, "[-c/--conf file] ");
	fprintf(stderr, "[-sp/--server-port port] ");
	fprintf(stderr, "[-mt/--max-threads number_of_threads] ");
	fprintf(stderr, "[-mq/--mq-key mq_key] ");
	fprintf(stderr, "[-td/--teams-dir directory] ");
	fprintf(stderr, "[-rdb/--rdb-file path] ");
	fprintf(stderr, "[-rfl/--rdb-free-list path]\n\n");

	fprintf(stderr, "\t-h --help\t\tDisplay this message\n");
	fprintf(stderr, "\t-c --conf\t\tTCA configuration file\n");
	fprintf(stderr, "\t-sp --server-port\tTCA server port\n");
	fprintf(stderr, "\t-mq --mq-key\t\tMessage queue key (for notification of new teams)\n");
	fprintf(stderr, "\t-mt --max-threads\tMaximum number of threads for TCP connections\n");
	fprintf(stderr, "\t-td --teams-dir\t\tDirectory where issued team public keys are stored\n");
	fprintf(stderr, "\t-rdb --rdb-file\t\tReceipt database file\n");
	fprintf(stderr, "\t-rfl --rdb-free-list\tReceipt database list of free positions file\n\n");
}

//##################################################################
int main(int argc, char** argv) {
	signal(SIGINT, term_handler);
	signal(SIGTERM, term_handler);

	struct sockaddr_in client_addr;
	int client_sockfd;
	//int counter = 0;
	int alen = 0;

	FILE *conf_file_fp = NULL;

	memset(&conf_file_path, 0, 256);
	memset(&teams_dir, 0, 256);
	memset(&rdb_file, 0, 256);
	memset(&rdb_free_list, 0, 256);

	int help = 0;

	CMD_ARG *cmd_line_args = NULL;
	CENTRY *conf_entries = NULL;
	unsigned char *valid_conf_names[] = {
		"server-port",
		"mq-key",
		"max-threads",
		"teams-dir",
		"rdb-file",
		"rdb-free-list"
		};

	unsigned char *valid_cmd_line_names[] ={
		"conf",
		"server-port",
		"mq-key",
		"max-threads",
		"teams-dir",
		"rdb-file",
		"rdb-free-list",
		"help"
		};

	unsigned char *cmd_line_switches[] = {"c", "sp", "mq", "mt", "td", "rdb", "rfl", "h"};

	unsigned short conf_var_types[] = {
		VTYPE_INT, //server port
		VTYPE_INT, //mq key
		VTYPE_INT, //max threads
		VTYPE_STRING, //teams directory
		VTYPE_STRING, //receipt db file
		VTYPE_STRING //receipts db free positions
		};

	unsigned short cmd_line_var_types[] = {
		VTYPE_STRING, //conffile
		VTYPE_INT, //server port
		VTYPE_INT, //mq key
		VTYPE_INT, //max threads
		VTYPE_STRING, //teams directory
		VTYPE_STRING, //receipt db file
		VTYPE_STRING, //receipts db free positions
		VTYPE_NOVAL //help
		};

	void *conf_variables[] = {
		&server_port,
		&mq_key,
		&max_threads,
		&teams_dir,
		&rdb_file,
		&rdb_free_list
		};

	void *cmd_line_variables[] = {
		&conf_file_path,
		&server_port,
		&mq_key,
		&max_threads,
		&teams_dir,
		&rdb_file,
		&rdb_free_list,
		&help
		};

	conf_entries = make_valid_entries(valid_conf_names, conf_variables, conf_var_types, 6);
	cmd_line_args = make_valid_args(valid_cmd_line_names, cmd_line_switches, cmd_line_variables, cmd_line_var_types, 8);

	if (argc < 2) {
		//no command line flags specified, so parse default conf file
		conf_file_fp = fopen(DEF_CONF_PATH, "r");
		if (!conf_file_fp) {
			//first check in /etc/pwnc.conf
			conf_file_fp = fopen(DEF_CONF_FNAME, "r");
			if (!conf_file_fp) {
				//if fopen failed, try to open pwnc.conf in the current path
				//if all else fails ---> exit
				fprintf(stderr, "P2PWNC TCA: Error opening configuration file\n\n");
				print_usage();
				exit(1);
			}
		}
		//default conffile opened ok, so parse it
		parse_conf_file(conf_entries, 6, conf_file_fp);
	}
	else {
		//parse command line
		//if there is a -c flag, ignore all other flags and check the specified conf file
		parse_cmd_line_args(argc, argv, cmd_line_args, 8);
		if (help) {
			//user gave the -h flag (probably among others...)
			print_usage();
			exit(0);
		}
		if (strlen(conf_file_path)) {
			//there was a -c flag, so ignore all other flags and check this conf file
			conf_file_fp = fopen(conf_file_path, "r");
			if (!conf_file_fp) {
				fprintf(stderr, "P2PWNC TCA: Warning: Failed to open %s, trying defaults\n", conf_file_path);
				//try to open specified conffile - if it fails, try defaults again
				conf_file_fp = fopen(DEF_CONF_PATH, "r");
				if (!conf_file_fp) {
					//trying default conffile path
					conf_file_fp = fopen(DEF_CONF_FNAME, "r");
					if (!conf_file_fp) {
						//trying to open default conf file in current dir
						fprintf(stderr, "P2PWNC TCA: Error opening configuration file\n\n");
						print_usage();
						exit(1);
					}
				}
			}
			parse_conf_file(conf_entries, 6, conf_file_fp);
		}
		//otherwise: we are using cmd line arguments
	}

	//check given values -- if not given, use defaults
	if (!server_port) {
		server_port = PWNC_TCA_PORT;
	}

	if (!strlen(rdb_file)) {
		strncpy(rdb_file, RDB_FILE, strlen(RDB_FILE));
	}

	if (!strlen(rdb_free_list)) {
		strncpy(rdb_free_list, RDB_FREE_LIST, strlen(RDB_FREE_LIST));
	}

	if (!mq_key) {
		mq_key = DEF_MQ_KEY;
	}

	if (!max_threads) {
		//if max_thread=0 --> wrong number format or sth,
		//use default anyway...
		max_threads = DEF_POOL_SIZE;
	}

	if (!strlen(teams_dir)) {
		strncpy(teams_dir, TEAMS_DIR, strlen(TEAMS_DIR));
	}

	fprintf(stderr, "TCA configuration:\n");
	fprintf(stderr, "---------------------------\n");
	fprintf(stderr, "Listening to port: %d\n", server_port);
	fprintf(stderr, "Message queue key: %d\n", mq_key);
	fprintf(stderr, "Thread pool size: %d\n", max_threads);
	fprintf(stderr, "Team public keys directory: %s\n", teams_dir);
	fprintf(stderr, "Receipt database file: %s\n", rdb_file);
	fprintf(stderr, "Receipt database free positions file: %s\n\n", rdb_free_list);

	rdb_init(rdb_file, rdb_free_list);
	load_teams(teams_dir);
	rdb_load_data();

	//initialize openssl mt support
	init_ossl_threads();

	//seed prng
	init_prng();

	init_thread_pool(max_threads);
	pthread_mutex_init(&thread_pool_mutex, NULL);

#ifdef FIXED_IDS
//fixed ids
//the tca must be able to add/remove teams dynamically
//Another program is responsible for sending msgs to the tca process so that a thread
//responsible for adding a new team can add the new team to the hash table that stores teams

	//start thread
	pthread_create(&team_handler_thread, NULL, (void*)&team_handler, NULL);
#endif

	//Sockaddr Initialization...
	//TCA address where it listens for receipts
	memset(&tca_addr, 0, sizeof(struct sockaddr_in));
	tca_addr.sin_family = AF_INET;
	//Port the TCA listens for such messages
	tca_addr.sin_port = htons(server_port);
	tca_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	//Socket creation and binding
	tca_sockfd = socket(PF_INET, SOCK_STREAM, 0);
	if (tca_sockfd < 0) {
		fprintf(stderr, "tca: socket failed\n");
		term_handler(0);
	}

	if (bind (tca_sockfd, (struct sockaddr*)&tca_addr, sizeof(tca_addr)) <0 ) {
		fprintf(stderr, "tca: bind failed\n");
		term_handler(0);
	}

	//Specify a size of request queue
	if (listen(tca_sockfd, 50) < 0) {
		fprintf(stderr,"listen failed\n");
		term_handler(0);
	}

	alen = sizeof(client_addr);
	fprintf(stderr, "TCA running...\n");

	alen = sizeof(client_addr);

	while (1) {
		if (  (client_sockfd=accept(tca_sockfd, (struct sockaddr *)&client_addr, &alen)) < 0) {
			fprintf(stderr, "Accept failed\n");
			//exit (1);
		}
		start_new_thread((void*)&request_handler, client_sockfd);
	}

	return 0;
}
//##################################################################
